// DeaServ.cpp: Implementierungsdatei
//

#include "stdafx.h"
#include "DigitalEA.h"
#include "DeaServ.h"
#include "opc.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CServerConfig

CServerConfig::CServerConfig()
    {
    nTargetPort = 49153;
	fSmartBox = FALSE;
	fDisable = FALSE;
    wInputs = 12;
    wOutputs = 12;
    wInitMask = 0;
    nPollStep = 9;
    nId = 0;
    }

/////////////////////////////////////////////////////////////////////////////
// CDeaServ

int       CDeaServ::g_nCount = 0;
CDeaServ* CDeaServ::g_root = NULL;        // statische Membervariable definieren

CDeaServ::CDeaServ( LPCTSTR lpszHostAddress, CServerConfig &cfg ) 
    : m_evtRecv( FALSE, TRUE )  // manuelles Event
    {
    InitializeCriticalSection( &m_csSend );
    m_sock = INVALID_SOCKET;
    m_hAsync = NULL;
    m_nConnState = CS_IDLE;
    m_fTimeout = FALSE;
    m_nPollCount = 0;
    m_nLastRealError = 0;
    m_hParentWindow = NULL;
    memset( &m_drvState, 0, sizeof(m_drvState) ); 
    m_wInRegQuality = OPC_QUALITY_COMM_FAILURE;
    m_wDevStateQuality = OPC_QUALITY_COMM_FAILURE;
    m_wOutRegValid = 0;
    m_wOutRegKnown = 0;
    m_wOutRegOwned = cfg.wInitMask;

    // Parameter merken.
    m_strTarget = lpszHostAddress;
    m_cfg = cfg;

    // In die globale Liste einfgen (immer am Kopf, das ist am einfachsten).
	m_next = g_root;
	m_prev = NULL;
	if( m_next )
		m_next->m_prev = this;
	g_root = this;
    g_nCount++;
	}


CDeaServ::~CDeaServ()
    {
    if( m_sock != INVALID_SOCKET )
        closesocket( m_sock );
    if( m_hAsync != NULL )
        WSACancelAsyncRequest( m_hAsync );

    // Aus der globalen Liste ausklinken.
	if( m_prev )
		m_prev->m_next = m_next;
	else
		g_root = m_next;
	if (m_next)
		m_next->m_prev = m_prev;
    g_nCount--;
    }


//Die folgenden Zeilen nicht bearbeiten. Sie werden vom Klassen-Assistenten bentigt.
#if 0
BEGIN_MESSAGE_MAP(CDeaServ, CSocket)
	//{{AFX_MSG_MAP(CDeaServ)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()
#endif	// 0

/////////////////////////////////////////////////////////////////////////////
// Member-Funktionen CDeaServ 


// Verbindung trennen, evtl. ausstehende asynchrone Requests canceln,
// Signalqualitten zurcksetzen.
// Der optionale Parameter hngt davon ab, aus welchem Grund die Verbindung
// geschlossen wird, und bestimmt im wesentlichen, ob und wie bald automatischer
// Neuaufbau versucht wird.
// Standardwert CS_IDLE: kein automatischer Neuaufbau.

void CDeaServ::Disconnect( int nState )
    {
    if( m_hAsync )
        {
        CString str;
        int nResult = WSACancelAsyncRequest( m_hAsync );
        if( nResult == 0 )
            m_hAsync = NULL;
        else
            {
            str.LoadString( nResult );
            TRACE( "WSACancelAsyncRequest(): %s\n", (LPCTSTR)str );
            }
        }
    //shutdown( m_sock, 2 );
    closesocket( m_sock );
    m_sock = INVALID_SOCKET;
    if( nState > CS_IDLE )
        nState = CS_IDLE;   // alle Werte darber sind nur intern erlaubt
    m_nConnState = nState;

    // OPC-Signalqualitten zurcksetzen.
    m_wOutRegValid = 0;
    if( m_wInRegQuality == OPC_QUALITY_GOOD )
        m_wInRegQuality  = OPC_QUALITY_LAST_KNOWN;
    if( m_wDevStateQuality == OPC_QUALITY_GOOD )
        m_wDevStateQuality  = OPC_QUALITY_LAST_KNOWN;
    }


// Rckgabewert ist nur dann FALSE, wenn wir sofort einen Fehler bekommen,
// Normalfall TRUE steht fr "wei nicht, warte auf OnConnect".
// Parameter ist ein Fenster, das unsere asynchronen Benachrichtigungen
// auffangen und an uns weiterleiten wird. NULL ist nicht erlaubt.

BOOL CDeaServ::Activate( HWND hwnd )
    {
    int nResult;

    if( hwnd == NULL )
        return FALSE;
    
    m_hParentWindow = hwnd;
    m_dwResponseTicks = 0;
    m_fUpdated = TRUE;          // Aktualisierung der Antwortzeit zunchst unterdrcken

    if( m_cfg.fDisable )
        return FALSE;
    if( m_nConnState > CS_IDLE )
        {
        TRACE( "Hey, bin noch in Zustand %d!\n", m_nConnState );
        return TRUE;
        }

    m_fSecondChance = FALSE;
	m_sock = socket( AF_INET, SOCK_STREAM, 0 );
	if( m_sock == INVALID_SOCKET )
        {
        LocalGetLastError();
        TRACE( "socket: Fehler %ld\n", m_nLastRealError );
        m_nConnState = CS_DEADEND;
        return FALSE;
        }
    TRACE( "Socket %d fuer %s\n", m_sock, m_strTarget );
    // Asynchronen Modus mit Windows-Benachrichtigungen aktivieren:
	WSAAsyncSelect( m_sock, m_hParentWindow, DM_SOCKEVENT,
    				FD_CONNECT | FD_READ | FD_WRITE | FD_CLOSE );

    // Zuerst versuchen, unseren Hostnamen als IP-Adresse zu interpretieren.
    // Wenn das geht, sparen wir uns einen Schritt.
    if( inet_addr( m_strTarget ) != INADDR_NONE )
        {
        SOCKADDR_IN sa;
    	sa.sin_family = AF_INET;
		sa.sin_port = htons( m_cfg.nTargetPort );
	    sa.sin_addr.s_addr = inet_addr( m_strTarget );

        m_nConnState = CS_RESOLVED;
	    nResult = connect( m_sock, (LPSOCKADDR)&sa, sizeof( SOCKADDR_IN ) );
	    if( nResult != 0 && LocalGetLastError()!= WSAEWOULDBLOCK )
            {
            TRACE( "connect: Fehler %ld\n", m_nLastRealError );
            Disconnect( CS_DEADEND );
            return FALSE;
            }
        }
    else
        {
        m_nConnState = CS_STARTED;
        m_hAsync = WSAAsyncGetHostByName( m_hParentWindow, DM_HOSTRESOLVED,
            m_strTarget, m_bufHostEnt, sizeof( m_bufHostEnt ) );
        if( m_hAsync == NULL )
            {
            LocalGetLastError();
            TRACE( "WSAAsyncGetHostByName(): Fehler %ld\n", m_nLastRealError );   
            Disconnect( CS_DEADEND );
            return FALSE;
            }
        }
        
    return TRUE;
    }


// OnHostResolved() wird von unserem freundlichen Fenster aufgerufen, wenn es
// die DM_HOSTRESOLVED-Nachricht bekommen hat.

void CDeaServ::OnHostResolved( LPARAM lParam )
    {
	LPHOSTENT lpHostEnt = (LPHOSTENT)m_bufHostEnt;
    SOCKADDR_IN sa;
    int nResult = WSAGETASYNCERROR( lParam );

    // Den asynchronen Aufruf als erledigt kennzeichnen. 
    m_hAsync = NULL;
    if( nResult != 0 )
        {
        m_nLastRealError = nResult;
        TRACE( "OnHostResolved: Fehler %d\n", m_nLastRealError );
        Disconnect( CS_DEADEND );
        return;
        }

    m_nConnState = CS_RESOLVED;
    sa.sin_family = AF_INET;
	sa.sin_port = htons( m_cfg.nTargetPort );
	sa.sin_addr = *((LPIN_ADDR)*lpHostEnt->h_addr_list);

	nResult = connect( m_sock, (LPSOCKADDR)&sa, sizeof( SOCKADDR_IN ) );
	if( nResult != 0 && LocalGetLastError()!= WSAEWOULDBLOCK )
        {
        TRACE( "connect: Fehler %ld\n", m_nLastRealError );
        Disconnect( CS_DEADEND );
        }
    }



// Von den Nachrichten ber Verbindungsaufbau interessieren uns hier nur die
// Fehlschlge. Erfolgreicher Verbindungsaufbau lst gleichzeitig ja eine
// OnSend()-Benachrichtigung aus.

void CDeaServ::OnConnect( int nErrorCode )
    {
    // Diese Zuweisung nehmen wir ausnahmsweise auch dann vor, wenn kein wirklicher
    // Fehler vorliegt. Anschaulich gesagt: erfolgreicher Verbindungsaufbau lscht 
    // den Fehlerspeicher.
    m_nLastRealError = nErrorCode;
    if( nErrorCode != 0 )
        {
        TRACE( "Verbindungsaufbau mit Code %d fehlgeschlagen\n", m_nLastRealError );
        Disconnect( CS_REJECTED );
        if( nErrorCode == WSAECONNREFUSED )
            m_nPollCount = 50;  // 5 Sek. Pause nach abgewiesener Verbindung
        }
    }


// OnSend() wird einmal aufgerufen, direkt nachdem die Verbindung aufgebaut
// wurde. Danach nur noch, wenn ein Send()-Aufruf mit WSAEWOULDBLOCK
// fehlgeschlagen ist, wenn also ein neues Send() versucht wurde, bevor
// WinSock bereit war, neue Daten anzunehmen. Und das wird relativ selten
// passieren: "Bereit" setzt nicht voraus, da die Daten des letzten Aufrufs
// schon quittiert wurden, sondern nur, da Winsock noch Platz in seinem
// lokalen Puffer hat!
// Sinnvolles Verhalten in der OnSend-Routine wird also sein, in einer
// Schleife so lange Daten zu senden, bis man abgewiesen wird, oder natrlich
// (was wahrscheinlicher ist) keine mehr zu senden hat.

void CDeaServ::OnSend( int nErrorCode )
    {
    CCritical crit( &m_csSend );
    int nNext;    // nchster Wert fr m_nConnState, *falls* Send() klappt
    int nResult;
    BOOL fSend;

    if( nErrorCode != 0 )
        {
        if( m_nConnState != CS_READY )
            Disconnect( CS_REJECTED );
        m_nLastRealError = nErrorCode;
        TRACE( "OnSend mit Fehlercode %d aufgerufen\n", m_nLastRealError );
        return;
        }

    m_fTimeout = FALSE;
    if( m_nConnState == CS_RESOLVED )
        m_nConnState = CS_CONNECTED;
    
    while( TRUE )
        {
        fSend = FALSE;
        switch( m_nConnState )
            {
            case CS_CONNECTED:
                // Nach erfolgreichem Verbindungsaufbau sofort als erstes die
                // Ausgnge wiederherstellen - jedenfalls die, die uns "gehren".
                _INIT_EA_DRIVER( m_eaDrvOut, SET_BIT );
                m_eaDrvOut.sb.set_bits = m_wOutRegOwned;
                m_eaDrvOut.sb.value    = m_drvState.output_reg;
                fSend = TRUE;
                nNext = CS_RESTORED;
                break;
            case CS_RESTORED:
                // Prima, Ausgnge wiederherstellen hat anscheinend geklappt:
                CoFileTimeNow( &m_ftOutputReg );
                m_wOutRegKnown =
                m_wOutRegValid = m_wOutRegOwned;
                // Die Eingangsvariablen abonnieren.
                _INIT_EA_DRIVER( m_eaDrvOut, SEND_MODE );
                m_eaDrvOut.sm.EA_bits = 0xfff;   // alle Bits berwachen
                m_eaDrvOut.sm.intervall = 0;     // aber keine Updates nach Zeitraster 
                fSend = TRUE;
                nNext = m_cfg.fSmartBox ? CS_SUBSCRIBED : CS_READY;
                break;
            case CS_SUBSCRIBED:
                // Nur fr Gerte mit Mefunktion: Alarme anfordern.
                // Interessant sind Spannungsberwachung (gibt oft noch eine 
                // Meldung kurz vor komplettem Spannungsausfall) und Kurzschlu.
                _INIT_EA_DRIVER( m_eaDrvOut, SET_ALARM );
                m_eaDrvOut.sa.VCC = TRUE;
                m_eaDrvOut.sa.kurzschluss = 0xfff;
                fSend = TRUE;
                nNext = CS_READY;
                break;
            case CS_READY:
                PostMessage( m_hParentWindow, DM_UPDATED, 0, (LPARAM)this );
                break;
            default:
                TRACE( "Huh? Zustand %d?", m_nConnState );
            }
        if( fSend )
            {
            nResult = send( m_sock, (char *)&m_eaDrvOut, m_eaDrvOut.struct_length, 0 );
            if( nResult == m_eaDrvOut.struct_length )
                {
                m_nConnState = nNext;
                TRACE( "Verbindung: %d\n", m_nConnState );
                continue;
                }
            else if( nResult == SOCKET_ERROR )
                {
                LocalGetLastError();
                Disconnect( CS_REJECTED );
                TRACE( "Send: Fehler %ld\n", m_nLastRealError );
                }
            else
                // Wird hoffentlich nie vorkommen, denn eine korrekte 
                // Behandlung dieses Falls wre ziemlich kompliziert.
                TRACE( "Hoppla: nur %d von %d Byte gesendet\n",
                    nResult, m_eaDrvOut.struct_length );
            }
        break;
        }
    }


// OnTimer() sollte alle 100 ms aufgerufen werden.
// Sofern die Verbindung bereit ist, wird es dann alle (m_cfg.nPollStep)
// Aufrufe eine Statusabfrage an den E/A-Server abschicken. Wenn festgestellt
// wird, da noch keine Antwort auf die zuletzt gesendete Statusabfrage
// vorliegt, wird die Verbindung getrennt und zum Neuaufbau vorgemerkt.
// Ein solcher Neuaufbau wird ebenfalls von hier aus veranlat, aber erst bei
// einem spteren Aufruf, um die Nachrichtenverarbeitung nicht zu verwirren.

void CDeaServ::OnTimer()
    {
    CCritical crit( &m_csSend );
    int nResult;

    // Die meisten Aufrufe stellen nur den internen Modulo-Zhler weiter.
    if( --m_nPollCount > 0 ) 
        return;
    m_nPollCount = m_cfg.nPollStep;

    // Zhler ist bergelaufen. Was gibt es zu tun?
    switch( m_nConnState )
        {
        case CS_READY:          // Normalbetrieb:
            break;              // Gerteabfrage, s. unten.
        case CS_IDLE:           // Objekt ist auf Benutzerwunsch inaktiv 
        case CS_DEADEND:        // bzw. auf nicht behebbaren Fehler gestoen:
            return;             // Nichts tun.
        case CS_DOWN:           // Verbindung wurde unplanmig getrennt
        case CS_REJECTED:       // bzw. Verbindungsaufbau ist fehlgeschlagen:
            Activate( m_hParentWindow );    // Neuaufbau versuchen.
            return;
        default:                // Zwischenzustnde beim Verbindungsaufbau:
            return;             // Nichts tun.
        }

    // Timeoutprfung.
    // Bei Gerten mit Mefunktion, die uns noch nie eine DRV_STATE-Struktur
    // geschickt haben, versuchen wir erst noch, sie als Gerte ohne Mefunktion
    // anzusprechen, bevor wir sie fr tot erklren.
    if( !m_fTimeout )
        m_fSecondChance = FALSE;
    if( m_fTimeout && m_cfg.fSmartBox && m_wDevStateQuality == OPC_QUALITY_COMM_FAILURE
        && !m_fSecondChance )
        {    
        m_fTimeout = FALSE;
        m_fSecondChance = TRUE;
        }

    if( m_fTimeout )
        {
        TRACE( "Timeout! Verbindung wird resettet!\n" );
        Disconnect( CS_DOWN );
        m_nPollCount = 1;       // Neuaufbau sofort beim nchsten Mal versuchen
        return;
        }
    
    // Gertezustand abfragen.
    if( m_cfg.fSmartBox && !m_fSecondChance )
        // Bei Gerten mit Mefunktion den kompletten Status abfragen.
        INIT_EA_DRIVER( m_eaDrvOut, REQ_STATE );
    else
        // Bei Gerten ohne Mefunktion nur die Eingnge aktualisieren lassen.
        INIT_EA_DRIVER( m_eaDrvOut, READ_REG );

    m_fUpdated = FALSE;
    CoFileTimeNow( &m_ftQuery );
    nResult = send( m_sock, (char *)&m_eaDrvOut, m_eaDrvOut.struct_length, 0 );
    if( nResult == SOCKET_ERROR )
        {
        LocalGetLastError();
        TRACE( "Fehler %d in OnTimer()\n", m_nLastRealError );
        }
    m_fTimeout = TRUE;      // wird bei Empfang einer Antwort zurckgesetzt
    }



// Verschickt eine dem Gertetyp entsprechende Statusabfrage. Gibt FALSE zurck,
// wenn das nicht mglich ist. 

BOOL CDeaServ::UpdateNow()
    {
    CCritical crit( &m_csSend );
    if( m_nConnState != CS_READY )
        return FALSE;

    if( m_cfg.fSmartBox )
        // Bei Gerten mit Mefunktion den kompletten Status abfragen.
        INIT_EA_DRIVER( m_eaDrvOut, REQ_STATE );
    else
        // Bei Gerten ohne Mefunktion nur die Eingnge aktualisieren lassen.
        INIT_EA_DRIVER( m_eaDrvOut, READ_REG );

    m_evtRecv.ResetEvent();
    m_fUpdated = FALSE;
    CoFileTimeNow( &m_ftQuery );
    if( send( m_sock, (char *)&m_eaDrvOut, m_eaDrvOut.struct_length, 0 ) == SOCKET_ERROR )
        {
        LocalGetLastError();
        TRACE( "Fehler %d in UpdateNow()\n", m_nLastRealError );
        return FALSE;
        }

    return TRUE;
    }



// OnReceive() wird aufgerufen, wenn neue Daten eingetroffen sind. Vorsicht:
// Normalerweise handelt es sich dabei zwar um genau eine Datenstruktur,
// aber es knnten durchaus mal mehrere miteinander verschmelzen, z. B. weil
// wir zu langsam waren. Wir lesen darum so lange, bis keine Daten mehr da
// sind (bis Receive() unseren Puffer nicht mehr ganz fllt bzw. einfach 
// WSAEWOULDBLOCK liefert).

void CDeaServ::OnReceive( int nErrorCode )
    {
    int nDemand, nReceived;

    while( TRUE )
        {
        // Erst den Header empfangen und die Anzahl angekndigter Bytes anschauen.
        nReceived = recv( m_sock, (char *)&m_eaDrvIn, SIZE_EA_HEADER, 0 );
        if( nReceived == SOCKET_ERROR  )
            break;              // zur Fehlerbehandlung
        m_fTimeout = FALSE;     // wir haben tatschlich Daten empfangen
        if( nReceived < SIZE_EA_HEADER )
            {
            TRACE( "Receive: %d Byte Paket, unvollstaendiger Header\n", nReceived );
            return;
            }
        // Den Rest des Pakets empfangen.
        nDemand = m_eaDrvIn.struct_length;
        if( nDemand > sizeof( m_eaDrvIn ) )
            {                           // evtl. zu groes Paket abschneiden, gar nicht gut
            TRACE( "Receive: %d Byte Struktur, zu gross fuer uns\n", nDemand );
            nDemand = sizeof( m_eaDrvIn );
            }
        nDemand -= SIZE_EA_HEADER;      // bereits empfangenen Header abziehen
        nReceived = recv( m_sock, (char *)&m_eaDrvIn.wr, nDemand, 0 );
        if( nReceived == SOCKET_ERROR  )
            break;
        if( nReceived < nDemand )
            {
            TRACE( "Receive: unvollstaendige Struktur, %d von %d Byte fehlen\n", 
                nDemand - nReceived, m_eaDrvIn.struct_length );
            return;
            }
        // Pakettyp untersuchen.
        switch( m_eaDrvIn.struct_typ )
            {
            case ID_WRITE_REG:
                if( m_eaDrvIn.struct_length != SIZE_WRITE_REG 
                        + 2*(m_eaDrvIn.wr.word_anz - 1) )
                    TRACE( "Falsche Laenge fr WRITE_REG: %d Byte bei %d Words\n",
                        m_eaDrvIn.struct_length, m_eaDrvIn.wr.word_anz );
                else
                    {
                    m_drvState.input_reg = m_eaDrvIn.wr.regs[ 0 ];
                    //TRACE( "Eingaenge: %03x\n", m_drvState.input_reg );
                    CoFileTimeNow( &m_ftInputReg );
                    m_wInRegQuality = OPC_QUALITY_GOOD;
                    if( !m_cfg.fSmartBox && !m_fUpdated )
                        {
                        m_dwResponseTicks = m_ftInputReg.dwLowDateTime
                                          - m_ftQuery.dwLowDateTime;
                        //TRACE( "Box%d: %ld ms\n", m_cfg.nId, m_dwResponseTicks / 10000 );
                        m_fUpdated = TRUE;
                        if( !m_evtRecv.Lock( 0 ) )
                            TRACE( "SetEvent()\n" );
                        m_evtRecv.SetEvent();
                        }
                    }
                break;
            case ID_DRV_STATE:
                if( m_eaDrvIn.struct_length != SIZE_DRV_STATE )
                    TRACE( "Falsche Laenge fr DRV_STATE: %d Byte\n", m_eaDrvIn.struct_length );
                else
                    {
                    m_drvState = m_eaDrvIn.ds;
                    /*
                    TRACE( "Eingaenge: %03x, %d V, %d C", 
                        m_drvState.input_reg, m_drvState.VCC, m_drvState.temperatur );
                    TRACE( ", Ausgaenge: %03x (KS %03x)\n", 
                        m_drvState.output_reg, m_drvState.kurzschluss );
                        */
                    CoFileTimeNow( &m_ftDevState );
                    CoFileTimeNow( &m_ftInputReg );
                    CoFileTimeNow( &m_ftOutputReg );
                    m_wDevStateQuality = OPC_QUALITY_GOOD;
                    m_wInRegQuality    = OPC_QUALITY_GOOD;
                    m_wOutRegKnown =
                    m_wOutRegValid = 0xfff;
                    if( !m_fUpdated )
                        {
                        m_dwResponseTicks = m_ftInputReg.dwLowDateTime
                                          - m_ftQuery.dwLowDateTime;
                        //TRACE( "Box%d: %ld ms\n", m_cfg.nId, m_dwResponseTicks / 10000 );
                        m_fUpdated = TRUE;
                        if( !m_evtRecv.Lock( 0 ) )
                            TRACE( "SetEvent()\n" );
                        m_evtRecv.SetEvent();
                        }
                    }
                break;
            case ID_ALARM_STATE:
                if( m_eaDrvIn.struct_length != SIZE_ALARM_STATE )
                    TRACE( "Falsche Laenge fr ALARM_STATE: %d Byte\n", m_eaDrvIn.struct_length );
                else
                    {
                    TRACE( "ALARM: " );
                    if( m_eaDrvIn.as.kurzschluss )
                        {
                        assert( (void *)&m_eaDrvIn.as.reset == (void *)&m_eaDrvIn.as );
                        memcpy( &m_drvState.reset, &m_eaDrvIn.as.reset, sizeof( ALARM_STATE ) );
                        TRACE( "Kurzschluss: %03x ", m_drvState.kurzschluss );
                        CoFileTimeNow( &m_ftDevState );
                        m_wDevStateQuality = OPC_QUALITY_GOOD;
                        }
                    if( m_eaDrvIn.as.temperatur )
                        {
                        m_drvState.temperatur = m_eaDrvIn.as.temperatur;
                        TRACE( "%d C ", m_drvState.temperatur );
                        }
                    if( m_eaDrvIn.as.VCC )
                        {
                        m_drvState.VCC = m_eaDrvIn.as.VCC;
                        TRACE( "%d V ", m_drvState.VCC );
                        }
                    if( m_eaDrvIn.as.reset )
                        {
                        m_drvState.reset = m_eaDrvIn.as.reset;
                        TRACE( "Reset " );
                        }
                    TRACE( "\n" );
                    }
                break;
            default:
                TRACE( "Pakettyp %d, %d Byte\n", 
                    m_eaDrvIn.struct_typ, m_eaDrvIn.struct_length );
            }
        // Wir haben ein Paket verarbeitet:
        PostMessage( m_hParentWindow, DM_UPDATED, 0, (LPARAM)this );
        }

    // Normalerweise verlassen wird die Schleife mit WSAEWOULDBLOCK, wenn nmlich
    // keine zu lesenden Daten mehr da sind.
    if( LocalGetLastError() != WSAEWOULDBLOCK )
        TRACE( "Receive: Fehler %d\n", m_nLastRealError );
    }



void CDeaServ::OnClose( int nErrorCode )
    {
    if( nErrorCode == 0 )
        TRACE( "normales Verbindungsende\n" );
    else
        {
        TRACE( "Verbindungsende mit Code %d\n", nErrorCode );
        m_nLastRealError = nErrorCode;
        }
    }



// Den Kurzschlu-Alarm quittieren und damit die Ausgnge wieder einschalten

BOOL CDeaServ::ReEnable()
    {
    CCritical crit( &m_csSend );
    int nResult;

    _INIT_EA_DRIVER( m_eaDrvOut, CLEAR_ALARM );
    m_eaDrvOut.ca.kurzschluss = 0xfff;
    nResult = send( m_sock, (char *)&m_eaDrvOut, m_eaDrvOut.struct_length, 0 );
    if( nResult == SOCKET_ERROR )
        {
        LocalGetLastError();
        TRACE( "Fehler %d in ReEnable()\n", m_nLastRealError );
        return FALSE;
        }
    return TRUE;
    }


// Ausgangs-Bits setzen.

BOOL CDeaServ::WriteBits( WORD wMask, WORD wState )
    {
    CCritical crit( &m_csSend );
    int nResult;

    _INIT_EA_DRIVER( m_eaDrvOut, SET_BIT );
    m_eaDrvOut.sb.set_bits = wMask;
    m_eaDrvOut.sb.value    = wState & wMask;    // Maskierung wichtig, wg. Bug im E/A-Server
    nResult = send( m_sock, (char *)&m_eaDrvOut, m_eaDrvOut.struct_length, 0 );
    if( nResult == SOCKET_ERROR )
        {
        LocalGetLastError();
        TRACE( "Fehler %d in WriteBits()\n", m_nLastRealError );
        return FALSE;
        }
    m_drvState.output_reg &= ~wMask;
    m_drvState.output_reg |= (wMask & wState);
    CoFileTimeNow( &m_ftOutputReg );
    m_wOutRegValid |= wMask;
    m_wOutRegKnown |= wMask;
    m_wOutRegOwned |= wMask;
    return TRUE;
    }


int CDeaServ::LocalGetLastError()
    {
    int nError = WSAGetLastError();
    if( nError != 0 && nError != WSAEWOULDBLOCK )
        m_nLastRealError = nError;
    return nError;
    }


// Aus unseren diversen Zustandsflags die OPC-Signalqualitt einzelner Variablen
// ermitteln.
// Wre im Prinzip auch in CDeaTag::Update() ganz gut aufgehoben, geht aber 
// irgendwie doch zu tief in die Details der CDeaServ-Implementierung.

WORD CDeaServ::GetItemQuality( int nType, int nOffset )
    {
    if( m_cfg.fDisable )
        return OPC_QUALITY_NOT_CONNECTED;
    else if( m_nConnState == CS_DEADEND )
        // tritt typischerweise auf, wenn ein ungltiger Hostname angegeben wurde
        return OPC_QUALITY_CONFIG_ERROR;
    else switch( nType )
        {
        case DEA_OUTBIT:
            if( m_drvState.kurzschluss )
                return OPC_QUALITY_OUT_OF_SERVICE;
            else if( m_wOutRegValid & (1 << nOffset) )
                return OPC_QUALITY_GOOD;
            else if( m_wOutRegKnown & (1 << nOffset) )
                return OPC_QUALITY_LAST_KNOWN;
            else if( m_nConnState != CS_READY )
                return OPC_QUALITY_COMM_FAILURE;
            else
                return OPC_QUALITY_UNCERTAIN;
        case DEA_INBIT:
            return m_wInRegQuality;
        case DEA_LOAD:
            if( m_wDevStateQuality == OPC_QUALITY_GOOD && m_drvState.kurzschluss )
                return OPC_QUALITY_OUT_OF_SERVICE;
            return m_wDevStateQuality;
        default:
            return m_wDevStateQuality;
        }
    }



// Aufgrund der zuletzt ermittelten Antwortzeit unseres Servers und dem letzten
// Aktualisierungszeitpunkt dieses Itemtyps entscheiden, ob es fr "Read from
// DEVICE" lohnen wrde, tatschlich eine Statusanfrage abzuschicken.

BOOL CDeaServ::ItemIsStale( int nType, int nOffset )
    {
    FILETIME ftNow;
    DWORD dwAge;

    if( m_nConnState != CS_READY )      // Aktualisierung zwecklos
        return FALSE;
    CoFileTimeNow( &ftNow );
    switch( nType )
        {
        case DEA_INBIT:
            dwAge = ftNow.dwLowDateTime - m_ftInputReg.dwLowDateTime;
            break;
        case DEA_OUTBIT:
            if( !m_cfg.fSmartBox )
                // Aktualisierung ist zwecklos, wenn es um Ausgnge 
                // geht, die wir gar nicht lesen knnen!
                return FALSE;
            dwAge = ftNow.dwLowDateTime - m_ftOutputReg.dwLowDateTime;
            break;
        default:
            dwAge = ftNow.dwLowDateTime - m_ftDevState.dwLowDateTime;
        }

    //TRACE( "Age %ld, Response %ld", dwAge,  m_dwResponseTicks );
    return ( dwAge > m_dwResponseTicks );
    }


// Hilfsfunktion, die die vier Sorten von Winsock-Benachrichtigungen auf
// die zugehrigen Behandlungsroutinen verteilt.

void CDeaServ::OnSocketMessage( LPARAM lParam )
    {
    switch( WSAGETSELECTEVENT( lParam ) )
    	{
		case FD_CONNECT:
            OnConnect( WSAGETSELECTERROR( lParam ) );
			break;   
		case FD_READ:
            OnReceive( WSAGETSELECTERROR( lParam ) );
			break;
		case FD_CLOSE:
            OnClose( WSAGETSELECTERROR( lParam ) );
			break;
		case FD_WRITE:
            OnSend( WSAGETSELECTERROR( lParam ) );
			break;
        default:
            TRACE( "unbekannter Typ fr DM_SOCKEVENT!\n" );
        }
    }


// Statische Memberfunktion, die von der Fensterprozedur unseres bergeordneten 
// Fensters die von uns in Auftrag gegebenen WinSock-Benachrichtigungen
// weitergereicht bekommt. Das verursachende Objekt wird herausgesucht, und die
// letztendlichen Verarbeitung findet dann in einer gewhnlichen Memberfunktion
// statt.
// Auerdem sollte man das hier ca. alle 100 ms mit einer WM_TIMER Message
// aufrufen.

void CDeaServ::MessageHandler( UINT message, WPARAM wParam, LPARAM lParam )
    {
    CDeaServ *pdea;
    BOOL fFound = FALSE;

    switch( message )
        {
        case WM_TIMER:
            for( pdea = CDeaServ::g_root; pdea != NULL; pdea = pdea->m_next )
                pdea->OnTimer();
            break;
        case DM_HOSTRESOLVED:
            // Den E/A-Server, der diese Nachricht verursacht hat, finden wir 
            // ber das "asynchronous task handle", das er sich beim Aufruf
            // von WSAAsyncGetHostByName() gemerkt hat, und das hier als wParam
            // wieder auftaucht.
            for( pdea = CDeaServ::g_root; pdea != NULL; pdea = pdea->m_next )
                if( pdea->m_hAsync == (HANDLE)wParam )
                    {
                    pdea->OnHostResolved( lParam );
                    fFound = TRUE;
                    }
            if( !fFound )
                TRACE( "Kein Empfnger fr DM_HOSTRESOLVED!\n" );
            break;
        case DM_SOCKEVENT:
            // Auch hier mssen wir nach einem Empfnger erst suchen. Diesmal 
            // enthlt wParam die Socketnummer.
            for( pdea = CDeaServ::g_root; pdea != NULL; pdea = pdea->m_next )
                if( pdea->m_sock == (SOCKET)wParam )
        			{
	                pdea->OnSocketMessage( lParam );
                    fFound = TRUE;
                    }
            if( !fFound )
                TRACE( "Kein Empfaenger fuer DM_SOCKEVENT, Handle %d!\n", wParam );
            break;
        default:
            TRACE( "Nachricht %ld an CDeaServ weitergereicht ???\n" );
        }
    }
