#ifndef AFX_DEASERV_H__A510CB48_89B8_11D2_8F84_00802968F609__INCLUDED_
#define AFX_DEASERV_H__A510CB48_89B8_11D2_8F84_00802968F609__INCLUDED_

// DeaServ.h : Header-Datei
//

#include <afxmt.h>      // fr CEvent
#include "DeaPackets.h"

/////////////////////////////////////////////////////////////////////////////
// Befehlsziel CDeaServ 

// Zustnde fr m_nConnState:
enum { 
    CS_DOWN,        // Verbindung wurde unplanmig getrennt: so bald wie 
                    // mglich Neuaufbau versuchen.
    CS_REJECTED,    // Verbindungsversuch ist fehlgeschlagen: nach kurzer Pause
                    // nochmal versuchen. (Die Pause ist wichtig, um in Fllen
                    // von "Connection refused" unntige Netzlast zu vermeiden.)
    CS_DEADEND,     // Beim Verbindungsaufbau auf einen Fehler gestoen, der
                    // Neuversuche aussichtslos erscheinen lt (typ. ist
                    // das "ungltiger Hostname").
    CS_IDLE,        // noch keine Verbindung, bzw. Verbindung beendet
    CS_STARTED,     // Hostnamensauflsung begonnen
    CS_RESOLVED,    // Hostname aufgelst, Connect() aufgerufen
    CS_CONNECTED,   // OnConnect()-Callback ist eingetroffen
    CS_RESTORED,    // SET_BIT gesendet, SEND_MODE als nchstes
    CS_SUBSCRIBED,  // Eingangsregister abonniert, SET_ALARM senden
    CS_READY        // Verbindung steht, DEA-Server komplett initialisiert
    };

// Anwendungsspezifische Message-Codes, die wir an unser bergeordnetes
// Fenster schicken.
enum {
    DM_HOSTRESOLVED = WM_APP + 1,
    DM_SOCKEVENT,
    DM_UPDATED      // fr Benutzer interessante Zustandsnderung hat stattgefunden
    };

// Signaltypen (z. B. als Parameter fr GetItemQuality):
enum {
    DEA_INBIT,
    DEA_OUTBIT,
    DEA_LOAD,
    DEA_POWER,
    DEA_HEAT,
    DEA_SHUTDOWN
    };

// Parametersatz, wie er aus der Registry gelesen und dem Konstruktor bergeben
// wird. (Zumindest ein Teil davon: die Standard-Datentypen.)

class CServerConfig
    {
public:
    WORD wInputs;               // typ. 12 oder 2
    WORD wOutputs;
    WORD wInitMask;             // Ausgnge initialisieren, z. B. 0 oder 0xfff
    int  nPollStep;             // mal 100ms ergibt die Pollrate, typ. 1s
    UINT nTargetPort;           // normalerweise 49153
	BOOL fSmartBox;             // Gert mit Messfunktion?
	BOOL fDisable;              // Verbindungsaufbau unterdrcken?
    int  nId;                   // laufende Nummer, fr den OPC-Namen
    // Ein Standard-Konstruktor, um sinnvolle Default-Werte zu garantieren.
    CServerConfig();
    };


// Aus FSServer.h gestohlene Klasse (dort heit sie CSLock):
// Man legt ein Objekt dieses Typs an, um einen kritischen Abschnitt zu betreten
// und automatisch wieder zu verlassen, wenn der Gltigkeitsbereich des Objekts
// endet.

class CCritical
    {
public:
    CCritical( CRITICAL_SECTION *pLock )
        : m_pLock( pLock) { EnterCriticalSection( m_pLock ); }
    ~CCritical() { LeaveCriticalSection( m_pLock ); }
private:
    CRITICAL_SECTION *m_pLock;
    };



class CDeaServ 
{
// Implementierung einer globalen Liste ber alle CDeaServ-Objekte (wird
// jeweils im Konstruktor bzw. Destruktor aktualisiert).
public:
    static CDeaServ *g_root;    
    static int g_nCount;        
    CDeaServ *m_next;           // Zeiger zur Verkettung in der globalen Liste
    CDeaServ *m_prev;

// Fr Anwender interessante Variablen.
public:
	DRV_STATE m_drvState;       // aktuellstes Abbild unseres Servers
    // Zeitpunkte, wann bestimmte Aktualisierungen zuletzt stattgefunden haben:
	FILETIME m_ftDevState;      // DRV_STATE bzw. ALARM_STATE empfangen
	FILETIME m_ftInputReg;      // DRV_STATE bzw. WRITE_REG empfangen
	FILETIME m_ftOutputReg;     // DRV_STATE empfangen bzw. SET_BIT gesendet
    CString m_strTarget;        // IP-Adresse/Hostname,
    CServerConfig m_cfg;        //  restliche Konfig-Daten
	CEvent  m_evtRecv;          // Ermglicht blockierendes Warten auf TCP-Pakete.
    int     m_nConnState;       // Fortschritt beim Verbindungsaufbau, Aufrufer 
                                //  knnten evtl. auf CS_READY testen wollen.
	int     m_nLastRealError;   // Letzter Fehlercode von WinSock.

// Implementierungsdetails
private:
    SOCKET  m_sock;             // Socket-"Handle"
	HANDLE  m_hAsync;           // Referenz-Handle fr WSAAsyncGetHostByName
	char m_bufHostEnt[ MAXGETHOSTSTRUCT ];
	EA_DRIVER m_eaDrvIn;
	EA_DRIVER m_eaDrvOut;
    CRITICAL_SECTION m_csSend;  // Ausfllen und Verschicken der m_eaDrvOut-Struktur
                                // ist ein kritischer Abschnitt, denn dank der OPC-
                                // Callbacks knnten ohne Synchronisationsmanahmen
                                // tatschlich mehrere solcher Ablufe konkurrierend
                                // stattfinden.
    HWND    m_hParentWindow;    // Fenster, an das wir die oben definierten DM_XXX
                                // Nachrichten schicken, und auf dessen Mithilfe 
                                // wir insbesondere zur Verarbeitung der asynchronen
                                // Winsock-Meldungen angewiesen sind.
	int     m_nPollCount;       // luft Modulo (m_cfg.nPollStep), bei 0 wird gepollt
	BOOL    m_fTimeout;         // Wird bei jedem Aufruf von Pulse() auf TRUE
                                // gesetzt, auf FALSE innerhalb der diversen TCP-
                                // Callbacks. Eine echte Timeout-Situation liegt
                                // vor, wenn Pulse() bereits TRUE vorfindet.
	BOOL    m_fSecondChance;    // Gerte mit Mefunktion, die auf REQ_STATE nicht
                                // geantwortet haben, bekommen noch die Chance,
                                // auf READ_REG zu antworten, bevor wir sie fr
                                // ausgefallen erklren. (Fr den Fall, da nur
                                // ein Konfigurierungsfehler vorliegt...)
                                // FALSE bedeutet: Dem Gert steht noch eine
                                // zweite Chance zu.
    FILETIME m_ftQuery;         // Zeitpunkt der letzten Statusanfrage.
    BOOL     m_fUpdated;        // Wird FALSE beim Versenden einer Statusabfrage,
                                // TRUE beim Empfang einer entsprechenden Antwort
                                // (DRV_STATE, bzw. fr Server ohne Mefunktion: 
                                // WRITE_REG).
    DWORD    m_dwResponseTicks; // Daraus errechnete Antwortzeit, in 1/10s, 
                                // wird bei Verbindungsaufabu zu 0 initialisiert.
                                // Hintergedanke: Wenn der Wert im Prozeabbild,
                                // an dem wir interessiert sind, weniger alt ist 
                                // als diese Zeitspanne, wre direktes Lesen vom
                                // DEVICE kontraproduktiv und sollte einfach
                                // nicht ausgefhrt werden.
    // Die folgenden werden im Konstruktor auf OPC_QUALITY_COMM_FAILURE 
    // initialisiert, werden durch eintreffende Pakete OPC_QUALITY_GOOD und 
    // knnen von dort durch Timeouts auf OPC_QUALITY_LAST_KNOWN zurckfallen.
    WORD    m_wInRegQuality;
    WORD    m_wDevStateQuality;
    // Mit den Ausgangsbits ist es hnlich, nur komplizierter, da sie sich i. a.
    // nicht auslesen lassen. In vielen Fllen wissen wir darum nur von denen,
    // die wir selbst gesetzt haben, alle anderen sind OPC_QUALITY_UNCERTAIN. 
    // Mit denen, die wir selbst gesetzt haben, hat es noch etwas besonderes 
    // auf sich: Bei einem Verbindungsneuaufbau initialisieren wir sie mit den
    // zuletzt bekannten Werten.
    // Bitmaps:
    WORD    m_wOutRegValid;
    WORD    m_wOutRegKnown;
    WORD    m_wOutRegOwned;
    
// Operationen
public:
	CDeaServ( LPCTSTR lpszHostAddress, CServerConfig &cfg );
	virtual ~CDeaServ();
    BOOL Activate( HWND hwnd );
	void Disconnect( int nState = CS_IDLE );
	BOOL WriteBits( WORD wMask, WORD wState );
	BOOL ReEnable();
	WORD GetItemQuality( int nType, int nOffset );
	BOOL ItemIsStale( int nType, int nOffset );
	BOOL UpdateNow();
	static void MessageHandler( UINT message, WPARAM wParam, LPARAM lParam );
	int  LocalGetLastError();

// Implementierung
protected:
	void OnHostResolved( LPARAM lParam );
	void OnSocketMessage( LPARAM lParam );
	void OnConnect( int nErrorCode );
	void OnClose( int nErrorCode );
	void OnReceive( int nErrorCode );
	void OnSend( int nErrorCode );
	void OnTimer();

// berschreibungen
public:
	// Vom Klassen-Assistenten generierte virtuelle Funktionsberschreibungen
	//{{AFX_VIRTUAL(CDeaServ)
	//}}AFX_VIRTUAL

	// Generierte Nachrichtenzuordnungsfunktionen
	//{{AFX_MSG(CDeaServ)
		// HINWEIS - Der Klassen-Assistent fgt hier Member-Funktionen ein und entfernt diese.
	//}}AFX_MSG
};

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Developer Studio fgt zustzliche Deklarationen unmittelbar vor der vorhergehenden Zeile ein.

#endif // AFX_DEASERV_H__A510CB48_89B8_11D2_8F84_00802968F609__INCLUDED_
