/*++

Datei: DeaPackets.h

    Strukturdefinitionen zur Ansteuerung der 50xxx-E/A-Server

Erstellt: 20.07.98 (WN)

nderungen:

    02.12.98 - aus dem Test88xx-Projekt extrahiert und abgewandelt,
        zustzliche Strukuren, andere Strukturnamen, Init-Makros. (WN)

--*/

#ifndef INC_DEAPACKETS_H
#define INC_DEAPACKETS_H


//#include <windows.h>
#include <assert.h>


#define ID_READ_REG     0x0001
#define SIZE_READ_REG   8

#define ID_WRITE_REG    0x0008
#define SIZE_WRITE_REG  12      // hmmm..., ist in Wirklichkeit variabel 
typedef struct
    {
    WORD word_anz;              // beim Senden immer 1
    WORD regs[ 1 ];
    } WRITE_REG;

#define ID_SET_BIT      0x0009
#define SIZE_SET_BIT    12
typedef struct
    {
    WORD set_bits;              // zu ndernde Bits
    WORD value;                 // zuzuweisende Werte
    } SET_BIT;

#define ID_SEND_MODE    0x0010
#define SIZE_SEND_MODE  12
typedef struct
    {
    WORD EA_bits;               // Bits, deren nderung gemeldet werden soll
    WORD intervall;             // regelmige Updates, Zeiteinheit 100 ms
    } SEND_MODE;

#define ID_REQ_STATE    0x0020
#define SIZE_REQ_STATE  8

#define ID_DRV_STATE    0x0030
#define SIZE_DRV_STATE  58
typedef struct
    {
    WORD driver_id;             // z. Zt. immer 1 == "DEA mit Mefunktion"
    WORD input_reg;             // Bitmap E0..E11
    WORD output_reg;            // Bitmap A0..A11
    WORD reset;                 // !=0: Karte hat sich resettet
    CHAR temperatur;            // Grad Celsius im Gehuse
    CHAR VCC;                   // Versorgungsspannung in Volt
    WORD kurzschluss;           // Bitmap: 500mA berschritten an A0..A11 -> Abschaltung
    WORD AD_lim_exceed;         // Bitmap: programmierter Grenzwert A0..A11 berschritten
   UCHAR AD_mess[ 12 ];         // Ausgangsstrme, Einheit 3.3mA
    CHAR AD_low_limit[ 12 ];    // Stromberwachung,
    CHAR AD_high_limit[ 12 ];   //  Einheit 30mA, Werteberich 0..15 (!)
    } DRV_STATE;

#define ID_SET_ALARM    0x0080
#define SIZE_SET_ALARM  40
typedef struct
    {                           // Benachrichtigung, wenn:
    WORD reset;                 //  die Karte sich resettet hat
    CHAR temperatur;            //  Temperatur <0 oder >60 Grad
    CHAR VCC;                   //  VCC <9V oder >31V
    WORD kurzschluss;           //  Abschaltung A0..A11 wegen Kurzschlu (Flag)
    WORD AD_lim_exceed;         //  Stromberwachung A0..A11 berschritten (Bitmap) 
    CHAR AD_low_limit[ 12 ];    // Werte fr Stromberwachung,
    CHAR AD_high_limit[ 12 ];   //  Einheit 30mA, Werteberich 0..15 (!)
    } SET_ALARM;

#define ID_ALARM_STATE  0x0090
#define SIZE_ALARM_STATE    28
typedef struct
    {
    WORD reset;                 // !=0: Karte hat sich resettet
    CHAR temperatur;            // Grad Celsius
    CHAR VCC;                   // Versorgungsspannung in Volt
    WORD kurzschluss;           // Bitmap: 500mA berschritten an A0..A11 -> Abschaltung
    WORD AD_lim_exceed;         // Bitmap: programmierter Grenzwert A0..A11 berschritten
   UCHAR AD_mess[ 12 ];         // Ausgangsstrme, Einheit 3.3mA
    } ALARM_STATE;

#define ID_CLEAR_ALARM  0x00A0
#define SIZE_CLEAR_ALARM    14
typedef struct                  // Benachrichtigung besttigen und damit neue
    {                           //   ermglichen fr:
    char temperatur;            //  Temperatur <0 oder >60 Grad
    char VCC;                   //  VCC <9V oder >31V
    WORD kurzschluss;           //  Abschaltung A0..A11 wegen Kurzschlu
    WORD AD_lim_exceed;         //  Stromberwachung A0..A11 berschritten
    } CLEAR_ALARM;


#define SIZE_EA_HEADER       8
typedef struct
    {
    WORD send_sequence;         // immer 0, auer bei UDP mit Acknowledge-Verfahren
    WORD rec_sequence;          //  "
    WORD struct_typ;            // eine der ID_X_XX-Konstanten
    WORD struct_length;         // Lnge in Byte des gesamten Pakets
    union
        {
        ALARM_STATE as;
        CLEAR_ALARM ca;
        DRV_STATE   ds;
        SEND_MODE   sm;
        SET_ALARM   sa;
        SET_BIT     sb;
        WRITE_REG   wr;
        };
    } EA_DRIVER;

// Makro zum Initialisieren der EA_DRIVER-Struktur.
#define INIT_EA_DRIVER( ed, typ ) do { \
    memset( &ed, 0, SIZE_##typ ); \
    ed.struct_typ = ID_##typ; \
    ed.struct_length = SIZE_##typ; \
    } while( 0 )

// Variante, die auch gleich noch eine assert-Prfung der Strukturgre
// durchfhrt. (Es knnte immerhin sein, da der Compiler uns alle Komponenten
// auf WORD-Grenzen o. . ausgerichtet hat...)
#define _INIT_EA_DRIVER( ed, typ ) do { \
    memset( &ed, 0, SIZE_##typ ); \
    ed.struct_typ = ID_##typ; \
    ed.struct_length = SIZE_##typ; \
    /*TRACE( #typ ": %ld Byte\n", sizeof( typ ) );*/ \
    assert( ed.struct_length == sizeof( typ ) + SIZE_EA_HEADER ); \
    } while( 0 )

#endif  // INC_DEAPACKETS_H
