/*******************************************************************************/
/* serv_tcp.c										       (Win32 Application) */
/*                                                    Microsoft Visual C++ 5.0 */
/*******************************************************************************/
/* Beispielprogramm fr einen TCP-Server auf Port 7 (ECHO). Der Server akzep-  */
/* tiert mehrere Verbindungen und sendet das Echo der auf diesem Socket        */
/* empfangenen Daten an die Clients zurck, bis diese die Verbindung schlieen.*/
/*******************************************************************************/

#include <winsock.h>							// #include <windows.h>
#include <stdio.h>
#include "resource.h"

#define	ECHO_PORT	7
#define	MAX_SOCKETS	32
#define	BUFFER_SIZE	512							// maximal COM-Server Packet length

char    buffer[ BUFFER_SIZE ];
SOCKET  in_use[ MAX_SOCKETS ];			        // Client Handles
struct	timeval  seltime;
fd_set  readfds;

char    szBuffer[ 256 ];						// used for messages
char FAR* remoteIP;
	
void    read_and_echo( HWND, SOCKET );
void    make_new_echo( HWND, SOCKET );
void	ShowStatus( HWND, LPSTR, int );
long FAR PASCAL WTcpServProc( HWND, unsigned, WORD, LONG );
int WINAPI WinMain( HINSTANCE, HINSTANCE, LPSTR, int );


VOID wtcp_serv( HWND hWnd )
  {
	SOCKET MySock, x;
	struct sockaddr_in addr;
	MSG	   msg;

	for( x = 0; x < MAX_SOCKETS; x++ ) 
		in_use[ x ] = 0;
												// open socket, protocol TCP
	ShowStatus( hWnd, "socket()", ID_LISTING );
	if(( MySock = socket( AF_INET, SOCK_STREAM, 0 )) < 0 ) 
	  {	ShowStatus( hWnd, " failed.", ID_LISTING );
		return;
	  }
	addr.sin_family = AF_INET;
	addr.sin_port = htons( ECHO_PORT );
	addr.sin_addr.s_addr = 0;
												// bind socket to server port and IP-address
	ShowStatus( hWnd, "bind()", ID_LISTING );
	if( bind( MySock, (const struct sockaddr FAR *)&addr, sizeof(addr) ) < 0 ) 
	  {	( void )closesocket( MySock );
		ShowStatus( hWnd, " failed.", ID_LISTING );
		return;
	  }
												// listening for incoming connections 
	ShowStatus( hWnd, "listen()", ID_LISTING );
	if( listen( MySock, 5 ) < 0 ) 
	  {	( void )closesocket( MySock );
		ShowStatus( hWnd, " failed.", ID_LISTING );
		return;
	  }
	in_use[ 0 ] = MySock;
    EnableWindow( GetDlgItem( hWnd, ID_SERVER ), FALSE );
		
	for( ;; ) 
	  {
		FD_ZERO( (fd_set FAR *)&readfds );		// clear fd_count (struct fd_set)
		seltime.tv_sec = 0;
		seltime.tv_usec = 100;					// maximum time for select to wait
		
		for( x = 0; x < MAX_SOCKETS; x++ )		// set fd_set for all sockets
		  {	if( in_use[ x ] ) 
				FD_SET( in_use[x], (fd_set FAR *)&readfds );
		  }

		if( !select( MAX_SOCKETS,
					(fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
					(const struct timeval FAR *)&seltime )) 
		  {	
			if( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ))
			  {
	    		if( msg.message == WM_QUIT ) 
				  {	for( x = 0; x < MAX_SOCKETS; x++ )	// close all aktiv Clients
					   { if( in_use[ x ] && in_use[ x ] != MySock ) 
						     (void)closesocket( in_use[ x] );
					   }
				    (void)closesocket( MySock );
			    	return;
				  }
			    TranslateMessage( &msg );
			    DispatchMessage( &msg );
			  }
			continue;
		  }
		
		for( x = 0; x < MAX_SOCKETS; x++ ) 
		  {	if( FD_ISSET( in_use[x], (fd_set FAR *)&readfds ))
			  {	ShowStatus( hWnd, "select()", ID_LISTING );
				if( in_use[x] != MySock )				// data readable on present 
					read_and_echo( hWnd, in_use[x] );
				else 
					make_new_echo( hWnd, MySock );	// new connection request
			  }
		  }
	  }
  }


void read_and_echo( HWND hWnd, SOCKET iClient )	// Read and echo data
  {	int iLen, x;
	
	ShowStatus( hWnd, "recv()", ID_LISTING );
	if(( iLen = recv( iClient, (LPSTR) buffer, sizeof( buffer ), 0 )) <= 0 ) 
	  {	ShowStatus( hWnd, " failed.", ID_LISTING );
		for( x = 0; x < MAX_SOCKETS; x++ ) 
		     if( in_use[x] == iClient )
                 in_use[x] = 0;
		(void)closesocket( iClient );
		sprintf( szBuffer, "closesocket(%d)", iClient );
		ShowStatus( hWnd, szBuffer, ID_LISTING );
	  }
	else
	  { sprintf( szBuffer, " %d                 %d", iClient, iLen );
		ShowStatus( hWnd, szBuffer, ID_SOCKET );
		buffer[ iLen ] = 0;						// mark String end
		ShowStatus( hWnd, buffer, ID_DATA );	// write data to listbox
        ShowStatus( hWnd, "send()", ID_LISTING );
		if( send( iClient, (LPSTR) buffer, iLen, 0 ) < 0 )
		  { (void)closesocket( iClient );
      		for( x = 0; x < MAX_SOCKETS; x++ ) 
  		         if( in_use[x] == iClient )
                     in_use[x] = 0;
			ShowStatus( hWnd, "closesocket()", ID_LISTING );
  }	  }	  }


void make_new_echo( HWND hWnd, SOCKET MySock )	// accept new connection
  {	SOCKET  iClient;
	struct	sockaddr_in peer;
	int		x, peersize = sizeof( peer );
	    
	ShowStatus( hWnd, "accept()", ID_LISTING );
	if(( iClient = accept( MySock, 
						  (struct sockaddr FAR *)&peer, (LPINT)&peersize )) < 0 )
	  {	ShowStatus(hWnd, "WARNING: Active connections: accept", ID_LISTING);
		return;
	  }
	
	remoteIP = inet_ntoa( peer.sin_addr );		// IP-address -> Dot-Notation
	sprintf( szBuffer,"Open Connection from %s - Port %d (Handle %d)",
			 remoteIP, htons( peer.sin_port ), iClient );
	ShowStatus( hWnd, szBuffer, ID_LISTING );
    for( x = 0; x < MAX_SOCKETS; x++ ) 
		 if( in_use[x] == 0 )
           { in_use[x] = iClient;
             break;
           }
  }






long WINAPI WTcpServProc(
	HWND            hWnd,
	unsigned        message,
	WORD            wParam,
	LONG            lParam )

  {
	switch( message ) 
	  { 
		case WM_COMMAND:						// commands from menu items
			switch( wParam ) 
			  {	case ID_SERVER:					// button: Start Server
					wtcp_serv( hWnd );
				break;
				case ID_EXIT:					// button: Exit
				case IDCANCEL:					// button: close window
					DestroyWindow( hWnd );
				break;
			  }
		break;

		case WM_DESTROY:						// window closed
			PostQuitMessage( 0 );				// send WM_QUIT to Application
		break;

		default:
		break;
	  }
	return( 0 );
  }



void ShowStatus( HWND hDlg, LPSTR lpMsg, int mWnd )	// write state to listbox
  {
	int iEntries;
	
	
	SendMessage( GetDlgItem( hDlg, mWnd ),		// add new entry
							 LB_ADDSTRING, 
							 (WPARAM)-1, 
							 (LPARAM)lpMsg );
	iEntries = SendMessage( GetDlgItem( hDlg, mWnd ), 
										LB_GETCOUNT,
										0,                 
										0 );
	SendMessage( GetDlgItem( hDlg, mWnd ), 
							 LB_SETTOPINDEX,
							 iEntries-1,           
							 0 ); 
  }


int WINAPI WinMain(		
	HINSTANCE	hInstance,
	HINSTANCE	hPrevInstance,
	LPSTR		lpCmdLine,
	int			nShowCmd )
  {		
	WSADATA		wsadata;
												// init Winsock-DLL Version 1.1
	if( WSAStartup( MAKEWORD( 1,1 ), &wsadata ) == 0 )	
	  {
		DialogBox( hInstance, MAKEINTRESOURCE( IDD_WSSERVER ),
				   NULL, WTcpServProc );
		WSACleanup();
	  }
	else
		MessageBox( 0,  "Fehler beim Initialisieren der WINSOCK.DLL",
						"WSClient.EXE", 0 );
		
	return( 0 );
  }